/*
 * Copyright (c) 1998 Tal Davidson. All rights reserved.
 *
 * ASBeautifier.cpp 0.8.1 (part of astyle)
 * by Tal Davidson (davidsont@bigfoot.com)
 *
 * The astyle project and all files needed to compile it
 * are distributed under the "Artistic License" detailed below:
 *
 *
 *                           The ``Artistic License''
 *
 * Preamble
 * The intent of this document is to state the conditions under which a Package may
 * be copied, such that the Copyright Holder maintains some semblance of artistic
 * control over the development of the package, while giving the users of the
 * package the right to use and distribute the Package in a more-or-less customary
 * fashion, plus the right to make reasonable modifications.
 *
 * Definitions
 *     ``Package'' refers to the collection of files distributed by the Copyright
 *     Holder, and derivatives of that collection of files created through textual
 *     modification.
 *     ``Standard Version'' refers to such a Package if it has not been modified,
 *     or has been modified in accordance with the wishes of the Copyright Holder
 *     as specified below.
 *     ``Copyright Holder'' is whoever is named in the copyright or copyrights for
 *     the package.
 *     ``You'' is you, if you're thinking about copying or distributing this
 *     Package.
 *     ``Reasonable copying fee'' is whatever you can justify on the basis of media
 *     cost, duplication charges, time of people involved, and so on. (You will not
 *     be required to justify it to the Copyright Holder, but only to the computing
 *     community at large as a market that must bear the fee.)
 *     ``Freely Available'' means that no fee is charged for the item itself,
 *     though there may be fees involved in handling the item. It also means that
 *     recipients of the item may redistribute it under the same conditions they
 *     received it.
 *
 *     1. You may make and give away verbatim copies of the source form of the
 *        Standard Version of this Package without restriction, provided that you
 *        duplicate all of the original copyright notices and associated disclaimers.
 *     2. You may apply bug fixes, portability fixes and other modifications derived
 *        from the Public Domain or from the Copyright Holder. A Package modified in
 *        such a way shall still be considered the Standard Version.
 *     3. You may otherwise modify your copy of this Package in any way, provided that
 *        you insert a prominent notice in each changed file stating how and when you
 *        changed that file, and provided that you do at least ONE of the following:
 *         a. place your modifications in the Public Domain or otherwise make them
 *            Freely Available, such as by posting said modifications to Usenet or an
 *            equivalent medium, or placing the modifications on a major archive site
 *            such as uunet.uu.net, or by allowing the Copyright Holder to include
 *            your modifications in the Standard Version of the Package.
 *         b. use the modified Package only within your corporation or organization.
 *         c. rename any non-standard executables so the names do not conflict with
 *            standard executables, which must also be provided, and provide a
 *            separate manual page for each non-standard executable that clearly
 *            documents how it differs from the Standard Version.
 *         d. make other distribution arrangements with the Copyright Holder.
 *     4. You may distribute the programs of this Package in object code or executable
 *        form, provided that you do at least ONE of the following:
 *         a. distribute a Standard Version of the executables and library files,
 *            together with instructions (in the manual page or equivalent) on where
 *            to get the Standard Version.
 *         b. accompany the distribution with the machine-readable source of the
 *            Package with your modifications.
 *         c. give non-standard executables non-standard names, and clearly document
 *            the differences in manual pages (or equivalent), together with
 *            instructions on where to get the Standard Version.
 *         d. make other distribution arrangements with the Copyright Holder.
 *     5. You may charge a reasonable copying fee for any distribution of this
 *        Package. You may charge any fee you choose for support of this Package. You
 *        may not charge a fee for this Package itself. However, you may distribute
 *        this Package in aggregate with other (possibly commercial) programs as part
 *        of a larger (possibly commercial) software distribution provided that you do
 *        not advertise this Package as a product of your own. You may embed this
 *        Package's interpreter within an executable of yours (by linking); this shall
 *        be construed as a mere form of aggregation, provided that the complete
 *        Standard Version of the interpreter is so embedded.
 *     6. The scripts and library files supplied as input to or produced as output
 *        from the programs of this Package do not automatically fall under the
 *        copyright of this Package, but belong to whomever generated them, and may be
 *        sold commercially, and may be aggregated with this Package. If such scripts
 *        or library files are aggregated with this Package via the so-called "undump"
 *        or "unexec" methods of producing a binary executable image, then
 *        distribution of such an image shall neither be construed as a distribution
 *        of this Package nor shall it fall under the restrictions of Paragraphs 3 and
 *        4, provided that you do not represent such an executable image as a Standard
 *        Version of this Package.
 *     7. C subroutines (or comparably compiled subroutines in other languages)
 *        supplied by you and linked into this Package in order to emulate subroutines
 *        and variables of the language defined by this Package shall not be
 *        considered part of this Package, but are the equivalent of input as in
 *        Paragraph 6, provided these subroutines do not change the language in any
 *        way that would cause it to fail the regression tests for the language.
 *     8. Aggregation of this Package with a commercial distribution is always
 *        permitted provided that the use of this Package is embedded; that is, when
 *        no overt attempt is made to make this Package's interfaces visible to the
 *        end user of the commercial distribution. Such use shall not be construed as
 *        a distribution of this Package.
 *     9. The name of the Copyright Holder may not be used to endorse or promote
 *        products derived from this software without specific prior written
 *        permission.
 *        THIS PACKAGE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 *        WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 *        MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * The End
 */

#include <vector>
#include <string>
#include <algorithm>
#include <iostream>

#include "compiler_defines.h"
#include "ASBeautifier.h"

#define INIT_CONTAINER(container, value)     { if ( (container) != NULL ) delete (container); (container) = (value); }
#define DELETE_CONTAINER(container) { if ( (container) != NULL ) /*delete (container)*/ ; }

#ifdef USES_NAMESPACE
using namespace std;
#endif



    /*
    template<class C, class T>
    int indexOf(C &container, T *element)
    {
      C::iterator where;
      
      where= find(container.begin(), container.end(), element);
      if (where == container.end())
        return -1;
      else
        return where - container.begin();
    }
    */

    int indexOf(vector<string*> &container, string *element)
    {
        vector<string*>::iterator where;

        where= find(container.begin(), container.end(), element);
        if (where == container.end())
            return -1;
        else
            return where - container.begin();
    }


    string trim(string &str)
    {

        char start = 0;
        char end = str.length()-1;
         
	while (start < end && str[start] <= 0x20)
	  //while (start < end && (str[start] == ' ' || str[start] == '\t'))
            start++;

        while (start < end && str[end] <= 0x20)
	  //while (start < end && (str[end] == ' ' || str[end] == '\t'))
            end--;
         
        string returnStr(str, start, end+1);
        return returnStr;;
    }

#ifdef USES_NAMESPACE
namespace astyle {
#endif

    bool ASBeautifier::calledInitStatic = false;

    string ASBeautifier::IF = string("if");
    string ASBeautifier::ELSE = string ("else");
    string ASBeautifier::FOR = string("for");
    string ASBeautifier::DO = string("do");
    string ASBeautifier::WHILE = string("while");
    string ASBeautifier::SWITCH = string ("switch");
    string ASBeautifier::CASE = string ("case");
    string ASBeautifier::DEFAULT = string("default");
    string ASBeautifier::OTHERWISE = string ("otherwise");
    string ASBeautifier::CLASS = string("class");
    string ASBeautifier::STRUCT = string("struct");
    string ASBeautifier::UNION = string("union");
    string ASBeautifier::INTERFACE = string("interface");
    string ASBeautifier::NAMESPACE = string("namespace");
    string ASBeautifier::PUBLIC = string("public");
    string ASBeautifier::PROTECTED = string("protected");
    string ASBeautifier::PRIVATE = string("private");
    string ASBeautifier::STATIC = string("static");
    string ASBeautifier::SYNCHRONIZED = string("synchronized");
    string ASBeautifier::OPERATOR = string("operator");
    string ASBeautifier::TEMPLATE = string("template");
    string ASBeautifier::TRY = string("try");
    string ASBeautifier::CATCH = string("catch");
    string ASBeautifier::THROWS = string("throws");
    string ASBeautifier::OPEN_BRACKET = string("{");
    string ASBeautifier::CLOSE_BRACKET = string("}");
    string ASBeautifier::OPEN_LINE_COMMENT = string("//");
    string ASBeautifier::OPEN_COMMENT = string("/*");
    string ASBeautifier::CLOSE_COMMENT = string("*/");

    string ASBeautifier::ASSIGN = string("=");
    string ASBeautifier::PLUS_ASSIGN = string("+=");
    string ASBeautifier::MINUS_ASSIGN = string("-=");
    string ASBeautifier::MULT_ASSIGN = string("*=");
    string ASBeautifier::DIV_ASSIGN = string("/=");
    string ASBeautifier::MOD_ASSIGN = string("%=");
    string ASBeautifier::OR_ASSIGN = string("|=");
    string ASBeautifier::AND_ASSIGN = string("&=");
    string ASBeautifier::RETURN = string("return");

    string ASBeautifier::EQUAL = string("==");
    string ASBeautifier::PLUS_PLUS = string("++");
    string ASBeautifier::MINUS_MINUS = string("--");
    string ASBeautifier::NOT_EQUAL = string("!=");
    string ASBeautifier::GR_EQUAL = string(">=");
    string ASBeautifier::GR_GR = string(">>");
    string ASBeautifier::LS_EQUAL = string("<=");
    string ASBeautifier::LS_LS = string("<<");
    string ASBeautifier::ARROW = string("->");
    string ASBeautifier::AND = string("&&");
    string ASBeautifier::OR = string("||");



    vector<string*> ASBeautifier::headers;
    vector<string*> ASBeautifier::nonParenHeaders;
    vector<string*> ASBeautifier::preBlockStatements;
    vector<string*> ASBeautifier::assignmentOperators;
    vector<string*> ASBeautifier::nonAssignmentOperators;

    /**
     * initialize the static vars
     */
    void ASBeautifier::initStatic()
    {
        if (calledInitStatic)
            return;

        calledInitStatic = true;

        headers.push_back(&IF);
        headers.push_back(&ELSE);
        headers.push_back(&FOR);
        headers.push_back(&WHILE);
        headers.push_back(&DO);
        headers.push_back(&TRY);
        headers.push_back(&CATCH);
        headers.push_back(&SYNCHRONIZED);
        headers.push_back(&SWITCH);
        headers.push_back(&CASE);
        headers.push_back(&DEFAULT);
        headers.push_back(&STATIC);
        headers.push_back(&PUBLIC);
        headers.push_back(&PRIVATE);
        headers.push_back(&PROTECTED);
        headers.push_back(&OPERATOR);
        headers.push_back(&TEMPLATE);

        nonParenHeaders.push_back(&ELSE);
        nonParenHeaders.push_back(&DO);
        nonParenHeaders.push_back(&TRY);
        nonParenHeaders.push_back(&STATIC);
        nonParenHeaders.push_back(&CLASS);
        nonParenHeaders.push_back(&STRUCT);
        nonParenHeaders.push_back(&UNION);
         
        preBlockStatements.push_back(&CLASS);
        preBlockStatements.push_back(&STRUCT);
        preBlockStatements.push_back(&UNION);
        preBlockStatements.push_back(&INTERFACE);
        preBlockStatements.push_back(&NAMESPACE);
        preBlockStatements.push_back(&THROWS);
         
        assignmentOperators.push_back(&ASSIGN);
        assignmentOperators.push_back(&PLUS_ASSIGN);
        assignmentOperators.push_back(&MINUS_ASSIGN);
        assignmentOperators.push_back(&MULT_ASSIGN);
        assignmentOperators.push_back(&DIV_ASSIGN);
        assignmentOperators.push_back(&MOD_ASSIGN);
        assignmentOperators.push_back(&OR_ASSIGN);
        assignmentOperators.push_back(&AND_ASSIGN);
        assignmentOperators.push_back(&RETURN);

        nonAssignmentOperators.push_back(&EQUAL);
        nonAssignmentOperators.push_back(&PLUS_PLUS);
        nonAssignmentOperators.push_back(&MINUS_MINUS);
        nonAssignmentOperators.push_back(&NOT_EQUAL);
        nonAssignmentOperators.push_back(&GR_EQUAL);
        nonAssignmentOperators.push_back(&GR_GR);
        nonAssignmentOperators.push_back(&LS_EQUAL);
        nonAssignmentOperators.push_back(&LS_LS);
        nonAssignmentOperators.push_back(&ARROW);
        nonAssignmentOperators.push_back(&AND);
        nonAssignmentOperators.push_back(&OR);
    }

    ASBeautifier::ASBeautifier()
    {
        initStatic();

	headerStack  = NULL;
	tempStacks = NULL;
        blockParenDepthStack = NULL;
	blockStatementStack = NULL;
	parenStatementStack = NULL;
	bracketBlockStateStack = NULL;
	inStatementIndentStack = NULL;
	inStatementIndentStackSizeStack = NULL;
        parenIndentStack = NULL;

        init();
        setSpaceIndentation(4);
        setMaxInStatementIndentLength(40);
        setClassIndent(true);
        setSwitchIndent(true);
        setBracketIndent(false);
        setCStyle();
    }

    ASBeautifier::~ASBeautifier()
    {
        DELETE_CONTAINER( headerStack );
	DELETE_CONTAINER( tempStacks );
        DELETE_CONTAINER( blockParenDepthStack );
        DELETE_CONTAINER( blockStatementStack );
        DELETE_CONTAINER( parenStatementStack );
        DELETE_CONTAINER( bracketBlockStateStack );
        DELETE_CONTAINER( inStatementIndentStack );
        DELETE_CONTAINER( inStatementIndentStackSizeStack );
        DELETE_CONTAINER( parenIndentStack );
    }

    /**
     * initiate the ABeautifier.
     *
     * init() should be called every time a ABeautifier object is to start
     * beautifying a NEW source file.
     */


    void ASBeautifier::init()
    {
        INIT_CONTAINER( headerStack,  new vector<string*>() );
        INIT_CONTAINER( tempStacks, new vector< vector<string*>* >() );
        tempStacks->push_back(new vector<string*>);
         
        INIT_CONTAINER( blockParenDepthStack, new vector<int> );
        INIT_CONTAINER( blockStatementStack, new vector<bool> );
        INIT_CONTAINER( parenStatementStack, new vector<bool> );
         
        INIT_CONTAINER( bracketBlockStateStack, new vector<bool> );
        bracketBlockStateStack->push_back(true);

        INIT_CONTAINER( inStatementIndentStack, new vector<int> );
        INIT_CONTAINER( inStatementIndentStackSizeStack, new vector<int> );
        inStatementIndentStackSizeStack->push_back(0);
        INIT_CONTAINER( parenIndentStack, new vector<int> );
         
        //isSpecialChar = false;
        isInQuote = false;
        isInComment = false;
        isInStatement = false;
        isInCase = false;
        isInQuestion = false;
        isInClassHeader = false;
        isInClassHeaderTab = false;
        isInHeader = false;
        isInOperator = false;
        isInTemplate = false;

        templateDepth = 0;

        immediatelyPreviousAssignmentOp = NULL;

        parenDepth=0;  
        blockTabCount = 0;
        statementTabCount = -1;
        leadingWhiteSpaces = 0;

        prevNonSpaceCh = '{';
        currentNonSpaceCh = '{';

        prevNonLegalCh = '{';
        currentNonLegalCh = '{';
    }

    void ASBeautifier::setCStyle()
    {
        isCStyle = true;
    }

    void ASBeautifier::setJavaStyle()
    {
        isCStyle = false;
    }

    /**
     * ident using one tab per identation
     */
    void ASBeautifier::setTabIndentation()
    {
        indentString = "\t";
        indentLength = 4;
    }

    /**
     * ident a number of spaces for each identation.
     *
     * @param   length     number of spaces per indent.
     */
    void ASBeautifier::setSpaceIndentation(int length)
    {
        indentString=string(length, ' ');
         
        /*
          indentString.reserve(length);
          for (int i=0; i<length; i++)
            indentString.insert(4, ' ');
        */
        indentLength = length;
         
    }

    /**
     * set the maximum indentation between two lines in a multi-line statement.
     *
     * @param   max     maximum indentation length.
     */
    void ASBeautifier::setMaxInStatementIndentLength(int max)
    {
        maxInStatementIndent = max;
    }
     
    /**
     * set the state of the bracket indentation option. If true, brackets will 
     * be indented one additional indent.
     *
     * @param   state             state of option.
     */
    void ASBeautifier::setBracketIndent(bool state)
    {
        bracketIndent = state;
    }
     
    /**
     * set the state of the class indentation option. If true, C++ class
     * definitions will be indented one additional indent.
     *
     * @param   state             state of option.
     */
    void ASBeautifier::setClassIndent(bool state)
    {
        classIndent = state;
    }

     
    /**
     * set the state of the switch indentation option. If true, blocks of 'switch' 
     * statements will be indented one additional indent.
     *
     * @param   state             state of option.
     */
    void ASBeautifier::setSwitchIndent(bool state)
    {
        switchIndent = state;
    }

    /**
     * beautify a line of source code.
     *
     * every line of source code in a source code file should be sent
     * one after the other to the beautify method.
     */             
    string ASBeautifier::beautify(string &line)
    {
        bool isInLineComment = false; // true when the current character is in a // comment (such as this line ...)
        bool isInClass = false;
        bool isInSwitch = false;
	bool isSpecialChar = false;
        char ch = ' ';       // the current character
        char prevCh;         // previous char
        string outBuffer; // the newly idented line is bufferd here
        int tabCount = 0;  // number of indents before line
        string *lastLineHeader = NULL; // last header found within line
        bool closingBracketReached = false;
        int spaceTabCount = 0;
        bool usePreviousTabCount = false;
        int previousTabCount = 0;
        char tempCh;
        int headerStackSize = headerStack->size();
        bool isLineInStatement = isInStatement;
        bool shouldIndentBrackettedLine = true;

        int i;
         
        currentHeader = NULL;

        // handle and remove white spaces around the line:
        // If not in comment, first find out size of white space before line,
        // so that possible comments starting in the line continue in
        // relation to the preliminary white-space.
        if (!isInComment)
        {
            leadingWhiteSpaces = 0;
            while (leadingWhiteSpaces<line.length() && line[leadingWhiteSpaces] <= 0x20)
		   //         (line[leadingWhiteSpaces]==' ' ||
                   //            line[leadingWhiteSpaces]=='\t'))
                leadingWhiteSpaces++;
             
            line = trim(line); //<---
        }
        else
        {
            int trimSize;
            for (trimSize=0;
                    trimSize < line.length() && trimSize<leadingWhiteSpaces && line[trimSize]<= 0x20 ;
		   //(line[trimSize] == ' ' || line[trimSize] == '\t');
                    trimSize++)
                ;
            line = line.substr(trimSize); // same as substring?
        }
         

        if (line.length() == 0)
            return line;

        if (isCStyle && !isInComment && line[0] == '#')
            return line;

        if (!inStatementIndentStack->empty()) 
            spaceTabCount = inStatementIndentStack->back();	   
         

        // calculate preliminary indentation based on data from past lines
        for (i=0; i<headerStackSize; i++) 
        {	   
            if (!(i>0 && (*headerStack)[i-1] != &OPEN_BRACKET && (*headerStack)[i] == &OPEN_BRACKET))
                tabCount++;
             

            if (isCStyle && classIndent && i >= 1 &&
		  ((*headerStack)[i-1] == &CLASS
		   || (isCStyle && ((*headerStack)[i-1] == &STRUCT || (*headerStack)[i-1] == &UNION)))
		&& (*headerStack)[i] == &OPEN_BRACKET
		)
            {
                ++tabCount;
                isInClass = true;
            }
             
            // is the switchIndent option is on, indent switch statements an additional indent.
            else if (switchIndent && i > 1 &&
                    (*headerStack)[i-1] == &SWITCH &&
                    (*headerStack)[i] == &OPEN_BRACKET
                    )
            {
                ++tabCount;
                isInSwitch = true;
            }
        }

        if (isCStyle && isInClass && classIndent && headerStackSize >= 2 && 
                ( (*headerStack)[headerStackSize-2] == &CLASS
		  || (isCStyle && ((*headerStack)[headerStackSize-2] == &STRUCT
				   || (*headerStack)[headerStackSize-2] == &UNION ) ) )
	       && (*headerStack)[headerStackSize-1] == &OPEN_BRACKET && line[0] == '}')
            tabCount--;

        else if (isInSwitch && switchIndent && headerStackSize >= 2 && 
                (*headerStack)[headerStackSize-2] == &SWITCH &&
                (*headerStack)[headerStackSize-1] == &OPEN_BRACKET && line[0] == '}')
            tabCount--;
         
        if (isInClassHeader)
        {
            isInClassHeaderTab = true;
            tabCount += 2;
        }
        /*
        if (isInStatement)
          if (!headerStack.empty() && headerStack.back().compare(string("{")) != 0)
            tabCount--;
        */

         
        //cout<<"tabCount=" << tabCount << " headerStack.size()" << headerStack.size() << "\n";cout.flush();
        //for (vector<string>::iterator nn =headerStack.begin(); nn < headerStack.end(); nn++)
        //  cout<< *nn << " " ;
        //cout<<"\n";
        //cout << "tempStacks count = " << tempStacks.size() << "\n";
        //  cout.flush();

        // parse characters in the current line.

        for (i=0; i<line.length(); i++)
        {
     
	    tempCh = line[i];
	    if (tempCh < 0x20 || tempCh >= 0x7F)
	      continue;

            prevCh = ch;
            ch = tempCh;


             
            //if (ch == '\r')
            //    continue;

             
            outBuffer.append(1, ch);
             
            if (ch == ' ' || ch == '\t')
                continue;
             
            // handle special characters (i.e. backslash+character such as \n, \t, ...)
            if (isSpecialChar)
            {
                isSpecialChar = false;
                continue;
            }
            if (!(isInComment || isInLineComment) && line.COMPARE(i, 2, string("\\\\")) == 0)
            {
                outBuffer.append(1, '\\');
                i++;
                continue;
            }
            if (!(isInComment || isInLineComment) && ch=='\\')
            {
                isSpecialChar = true;
                continue;
            }
             
            // handle quotes (such as 'x' and "Hello Dolly")
            if (!(isInComment || isInLineComment) && (ch=='"' || ch=='\''))
                if (!isInQuote)
                {
                    quoteChar = ch;
                    isInQuote = true;
                }
                else if (quoteChar == ch)
                {
                    isInQuote = false;
                    isInStatement = true;
                    continue;
                }
            if (isInQuote)
                continue;
             
            //--

	    //-------------

            // handle comments
            if ( !(isInComment || isInLineComment) && line.COMPARE(i, 2, OPEN_LINE_COMMENT) == 0 )
            {
                isInLineComment = true;
                outBuffer.append(1, '/');
                i++;
                continue;
            }
            else if ( !(isInComment || isInLineComment) && line.COMPARE(i, 2, OPEN_COMMENT) == 0 )
            {
                isInComment = true;
                outBuffer.append(1, '*');
                i++;
                continue;
            }
            else if ( (isInComment || isInLineComment) && line.COMPARE(i, 2, CLOSE_COMMENT) == 0 )
            {
                isInComment = false;
                outBuffer.append(1, '/');
                i++;
                continue;
            }
             
            if (isInComment||isInLineComment)
                continue;


            //--
             
            // if we have reached this far then we are NOT in a comment or string of special character...

            prevNonSpaceCh = currentNonSpaceCh;
            currentNonSpaceCh = ch;
            if (!isLegalNameChar(ch) && ch != ',' && ch != ';' )
            {
                prevNonLegalCh = currentNonLegalCh;
                currentNonLegalCh = ch;
            }
            	

            if (isInHeader)
            {
                isInHeader = false;
	      
                currentHeader = headerStack->back();
            }
            else
                currentHeader = NULL; 


            if (isCStyle && isInTemplate
                    && (ch == '<' || ch == '>')
                    &&  findLegalHeader(line, i, nonAssignmentOperators) == -1);
            {
                if (ch == '<')
                {
                    ++templateDepth;
                }
                else if (ch == '>')
                {
                    if (--templateDepth <= 0)	
                    {	
                        if (isInTemplate)
                            ch = ';';
                        else
                            ch = 't';
                        isInTemplate = false;
                        templateDepth = 0;		
                    }		  		
                }
            }
             

             
            // handle parenthesies
            if (ch == '(' || ch == '[' || ch == ')' || ch == ']')
            {
                if (ch == '(' || ch == '[')
                {
                    if (parenDepth == 0)
                    {
                        parenStatementStack->push_back(isInStatement);
                        isInStatement = true;
                    }
                    parenDepth++;
                     
                    inStatementIndentStackSizeStack->push_back(inStatementIndentStack->size());
                    if (currentHeader != NULL)
                    {
                        inStatementIndentStack->push_back(indentLength*2 + spaceTabCount);
                        parenIndentStack->push_back(indentLength*2 + spaceTabCount);
                         
                    }
                    else			
                        registerInStatementIndent(line, i, spaceTabCount, isLineInStatement, true);
		    //lastParenInStatementIndent = isStatementIndentStack.back();
                }
                else if (ch == ')' || ch == ']')
                {     
                    parenDepth--;
                    if (parenDepth == 0)
                    {
                        isInStatement = parenStatementStack->back();
                        parenStatementStack->pop_back();
                        ch = ' ';
                    }
                    //--	    
                    if (!inStatementIndentStackSizeStack->empty())
                    {
                        int previousIndentStackSize = inStatementIndentStackSizeStack->back();
                        inStatementIndentStackSizeStack->pop_back();
                        while (previousIndentStackSize < inStatementIndentStack->size())
                            inStatementIndentStack->pop_back();
                         
                        if (!parenIndentStack->empty())
                        {
                            int poppedIndent = parenIndentStack->back();
                            parenIndentStack->pop_back();

                            if (i == 0)
                                spaceTabCount = poppedIndent;
                        }
                    }
                }

                continue;
            }
             

            if (ch == '{')
            { 
                bool isBlockOpener = false;
                 
                // first, check if '{' is a block-opener or an static-array opener
                isBlockOpener |= (prevNonSpaceCh == '{' &&
                                  bracketBlockStateStack->back());
                 
                isBlockOpener |= (prevNonSpaceCh == ')' || prevNonSpaceCh == ';');
                 
                isBlockOpener |= isInClassHeader;
                 
                isInClassHeader = false;
                 
                if (!isBlockOpener && currentHeader != NULL)
                {
                    for (int n=0; n < nonParenHeaders.size(); n++)
                        if (currentHeader == nonParenHeaders[n])
                        {
                            isBlockOpener = true;
                            break;
                        }
                }
                 
                bracketBlockStateStack->push_back(isBlockOpener);
                 
                if (!isBlockOpener)
                {  
                    if (line.length() - i == getNextProgramCharDistance(line, i) &&
                            immediatelyPreviousAssignmentOp != NULL) // && !inStatementIndentStack.isEmpty() - actually not needed
                        inStatementIndentStack->pop_back();
                    inStatementIndentStackSizeStack->push_back(inStatementIndentStack->size());
                    registerInStatementIndent(line, i, spaceTabCount, isLineInStatement, true);
                    //parenIndentStack.push(new Integer(i+spaceTabCount));
                    parenDepth++;
                    if (i == 0)
                        shouldIndentBrackettedLine = false;
                     
                    continue;
                }
                 
                if (isInClassHeader)   
                    isInClassHeader = false;
                if (isInClassHeaderTab)
                {
                    isInClassHeaderTab = false;
                    tabCount -= 2;
                }     
                 
                blockParenDepthStack->push_back(parenDepth);
                blockStatementStack->push_back(isInStatement);
                 
                inStatementIndentStackSizeStack->push_back(inStatementIndentStack->size());
                 
                blockTabCount += isInStatement? 1 : 0;
                parenDepth = 0;
                isInStatement = false;
                 
                tempStacks->push_back(new vector<string*>);
                headerStack->push_back(&OPEN_BRACKET);
                lastLineHeader = &OPEN_BRACKET; // <------
                 
                continue;   
            }
             



            //check if a header has been reached
            if (prevCh == ' ')
            {
                bool isDoubleHeader = false;
                int h = findLegalHeader(line, i, headers);
                if (h > -1)
                {
                    // if we reached here, then this is a header...
                     
                    isInHeader = true;		    
                     
                     
                    vector<string*> *lastTempStack;
                    if (tempStacks->empty())
                        lastTempStack = NULL;
                    else
                        lastTempStack = tempStacks->back();
                     
                    // if a new block is opened, push a new stack into tempStacks to hold the
                    // future list of headers in the new block.
                    //if ("{".equals(headers[h]))
                    //    tempStacks.push(new Stack());
                     
                    // take care of the special case: 'else if (...)'
                    if (headers[h] == &IF && lastLineHeader == &ELSE)
                    {
                        spaceTabCount += indentLength; // to counter the opposite addition that occurs when the 'if' is registered below...  
                        headerStack->pop_back();
                    }
                     
                    // take care of 'else'
                    else if (headers[h] == &ELSE)
                    {
                        //string header;
                        if (lastTempStack != NULL)
                        {
                            int indexOfIf = indexOf(*lastTempStack, &IF); // <---
                            if (indexOfIf != -1)
                            {
                                // recreate the header list in headerStack up to the previous 'if'
                                // from the temporary snapshot stored in lastTempStack.
                                int restackSize = lastTempStack->size() - indexOfIf - 1;
                                for (int r=0; r<restackSize; r++)
                                {
                                    headerStack->push_back(lastTempStack->back());
                                    lastTempStack->pop_back();
                                }
                                if (!closingBracketReached)
                                    tabCount += restackSize;
                            }
                            /*
                             * If the above if is not true, i.e. no 'if' before the 'else',
                             * then nothing beautiful will come out of this...
                             * I should think about inserting an Exception here to notify the caller of this...
                             */
                             
                        }
                    }
                     
                    // check if 'while' closes a previous 'do'
                    else if (headers[h] == &WHILE)
                    {
                        //string &header;
                        if (lastTempStack != NULL)
                        {
                            int indexOfDo = indexOf(*lastTempStack, &DO); // <---------
                            if (indexOfDo != -1)
                            {
                                // recreate the header list in headerStack up to the previous 'do'
                                // from the temporary snapshot stored in lastTempStack.
                                int restackSize = lastTempStack->size() - indexOfDo - 1;
                                for (int r=0; r<restackSize; r++)
                                {
                                    headerStack->push_back(lastTempStack->back());
                                    lastTempStack->pop_back();
                                }
                                if (!closingBracketReached)
                                    tabCount += restackSize;
                            }
                        }
                    }
                    // check if 'catch' closes a previous 'try' or 'catch'
                    else if (headers[h] == &CATCH)
                    {
                        //string header;
                        if (lastTempStack != NULL)
                        {
                            int indexOfTry = indexOf(*lastTempStack, &TRY);
                            if (indexOfTry == -1)
                                indexOfTry = indexOf(*lastTempStack, &CATCH);
                            if (indexOfTry != -1)
                            {
                                // recreate the header list in headerStack up to the previous 'do'
                                // from the temporary snapshot stored in lastTempStack.
                                int restackSize = lastTempStack->size() - indexOfTry - 1;
                                for (int r=0; r<restackSize; r++)
                                {
                                    headerStack->push_back(lastTempStack->back());
                                    lastTempStack->pop_back();
                                }
                                //lastTempStack.pop();
                                //headerStack.push("try");
                                 
                                if (!closingBracketReached)
                                    tabCount += restackSize;
                            }
                        }
                    }
                    else if (headers[h] == &CASE)
                    {
                        isInCase = true;
                        --tabCount;
                    }
                    else if(headers[h] == &OTHERWISE)
                    {
                        if (isCStyle)
                        {
                            isInCase = true;
                            --tabCount;
                        }
                        else
                            isDoubleHeader = true;
                    }
                    else if(headers[h] == &DEFAULT)
                    {
                        if (!isCStyle)
                        {
                            isInCase = true;
                            --tabCount;
                        }
                        else
                            isDoubleHeader = true;
                    }
                    else if (headers[h] == &PUBLIC || headers[h] == &PROTECTED || headers[h] == &PRIVATE)
                    {
                        if (isCStyle)
                            --tabCount;
                        isDoubleHeader = true;
                    }	 
                    else if ((headers[h] == &STATIC || headers[h] == &SYNCHRONIZED) &&
                            !headerStack->empty() &&
                            (headerStack->back() == &STATIC || headerStack->back() == &SYNCHRONIZED))
                    {
                        isDoubleHeader = true;
                    }
                    else if (headers[h] == &OPERATOR)
                    {
                        if (isCStyle)
                            isInOperator = true;
                        isDoubleHeader = true;
                    }
                    else if (headers[h] == &TEMPLATE)
                    {
                        if (isCStyle)
                            isInTemplate = true;
                        isDoubleHeader = true;
                    }
                     
                    if (isDoubleHeader)
		      isInHeader = false;
		    else
                    {
                        spaceTabCount-=indentLength;
                        headerStack->push_back(headers[h]);
                        isInStatement = false;
                    }
                     
		
                    lastLineHeader = headers[h];
                     
                    outBuffer.append(headers[h]->substr(1)); // <--- substring(1)
                    i += headers[h]->length() - 1;
                     
                    //isInStatement = false;

                    //if (isInOperator)
                    continue;
                }
            }
             
            if (ch == '?')
                isInQuestion = true;
             

            // special handling of 'case' statements
            if (ch == ':')
            {
                if (line.length() > i+1 && line[i+1] == ':')
                {
                    ++i;
                    outBuffer.append(1, ':');
                    ch = ' ';
                    continue;
                }
                else if (isInQuestion)
                {
                    isInQuestion = false;
                }
		else if (isCStyle && prevNonSpaceCh == ')')
		{
		  isInClassHeader = true;
		  if (i==0)
		    tabCount += 2;
		}
                else
                {
                    currentNonSpaceCh = ';'; // so that brackets after the ':' will appear as block-openers
                    if (isInCase)
                    {
                        isInCase = false;
                        ch = ';'; // from here on, treat char as ';'
                    }
                }
            } 
             
            if ((ch == ';'  || (parenDepth>0 && ch == ','))  && !inStatementIndentStackSizeStack->empty())
                while (inStatementIndentStackSizeStack->back() + (parenDepth>0 ? 1 : 0)  < inStatementIndentStack->size())
                    inStatementIndentStack->pop_back();
             
             
            // handle ends of statements
            if ( (ch == ';' && parenDepth == 0) || ch == '}'/* || (ch == ',' && parenDepth == 0)*/)
            {
                if (ch == '}')
                {
                    // first check if this '}' closes a previous block, or a static array...
                    if (!bracketBlockStateStack->empty())
                    {
                        bool bracketBlockState = bracketBlockStateStack->back();
                        bracketBlockStateStack->pop_back();
                        if (!bracketBlockState)
                        {
                            if (!inStatementIndentStackSizeStack->empty())
                            {
                                int previousIndentStackSize = inStatementIndentStackSizeStack->back();
                                inStatementIndentStackSizeStack->pop_back();
                                while (previousIndentStackSize < inStatementIndentStack->size())
                                    inStatementIndentStack->pop_back();
                                parenDepth--;
                                if (i == 0)
                                    shouldIndentBrackettedLine = false;
                                 
                                if (!parenIndentStack->empty())
                                {
                                    int poppedIndent = parenIndentStack->back();
                                    parenIndentStack->pop_back();
                                    if (i == 0)
                                        spaceTabCount = poppedIndent;
                                }			    
                            }		 
                            continue;
                        }
                    }
                     
                    if(!inStatementIndentStackSizeStack->empty())
                        inStatementIndentStackSizeStack->pop_back();
                     
                    if (!blockParenDepthStack->empty())
                    {
                        parenDepth = blockParenDepthStack->back();
                        blockParenDepthStack->pop_back();
                        isInStatement = blockStatementStack->back();
                        blockStatementStack->pop_back();
                         
                        if (isInStatement)
                            blockTabCount--;
                    }
                     
                    closingBracketReached = true;
                    int headerPlace = indexOf(*headerStack, &OPEN_BRACKET); // <---
                    if (headerPlace != -1)
                    {	     
                        string *popped = headerStack->back();
                        while (popped != &OPEN_BRACKET)
                        {
                            headerStack->pop_back();
                            popped = headerStack->back();
                        }
                        headerStack->pop_back();
                         
                        if (!tempStacks->empty())
                        {
                            vector<string*> *temp =  tempStacks->back();
                            tempStacks->pop_back();
                            delete temp;
                        }
                    }
                     
		    
                    ch = ' '; // needed due to cases such as '}else{', so that headers ('else' tn tih case) will be identified... 
                    //}
                }
                 
                //else if (ch == ';' /* parenDepth == 0*/)
                //    while (((Integer) inStatementIndentStackSizeStack.peek()).intValue() < inStatementIndentStack.size())
                //	inStatementIndentStack.pop();

                /*
                 * Create a temporary snapshot of the current block's header-list in the
                 * uppermost inner stack in tempStacks, and clear the headerStack up to
                 * the begining of the block.
                 * Thus, the next future statement will think it comes one indent past
                 * the block's '{' unless it specifically checks for a companion-header
                 * (such as a previous 'if' for an 'else' header) within the tempStacks,
                 * and recreates the temporary snapshot by manipulating the tempStacks.
                 */
                //if (tempStacks.empty())
                //{ cout<<"tempStacks empty!!\n"; cout.flush();}
                if (!tempStacks->back()->empty())	  
                    while (!tempStacks->back()->empty())
                        tempStacks->back()->pop_back();
                //tempStacks.back()->reserve(0);
                while (!headerStack->empty() && headerStack->back() != &OPEN_BRACKET)
                {
                    tempStacks->back()->push_back(headerStack->back());
                    headerStack->pop_back();
                }
                 
                if (parenDepth == 0 && ch == ';')
                    isInStatement=false;

		isInClassHeader = false;


                //for (vector<string>::iterator nn =headerStack.begin(); nn < headerStack.end(); nn++)
                //    cout<< *nn << " " ;
                //  cout<<"\n";cout.flush();
                 
                continue;
            }
             
             
            if (prevCh == ' ' && !isInTemplate)
            {
                int headerNum = findLegalHeader(line, i, preBlockStatements);
                if (headerNum > -1)
                {
                    isInClassHeader = true;
                    outBuffer.append(preBlockStatements[headerNum]->substr(1)); //<--- substring(1)
                    i += preBlockStatements[headerNum]->length() - 1;
                    if (isCStyle)
                    {
                        headerStack->push_back(preBlockStatements[headerNum]);		  
                        //lastLineHeader = &preBlockStatements[headerNum];
                    }
                }
            }
             
             
             
            // PRECHECK if a '==' or '--' or '++' operator was reached.
            // If not, then register an indent IF an assignment operator was reached.
            // The precheck is important, so that statements such as 'i--==2' are not recognized
            // to have assignment operators (here, '-=') in them . . .
            immediatelyPreviousAssignmentOp = NULL;
            bool isNonAssingmentOperator = false;
            for (int n = 0; n < nonAssignmentOperators.size(); n++)
                if (line.COMPARE(i, nonAssignmentOperators[n]->length(), *(nonAssignmentOperators[n])) == 0)
                {	  
                    outBuffer.append(nonAssignmentOperators[n]->substr(1)); //<---
                    i++;
                    //  the above two lines do the same as the next since all listed  non Assignment operators are 2 chars long...
                    // if (nonAssignmentOperators[n].length() > 1)
                    //{
                    //   outBuffer.append(nonAssignmentOperators[n].substring(1));
                    //   i += nonAssignmentOperators[n].length() - 1;
                    //}
                    //
                    isNonAssingmentOperator = true;
                    break;    
                }
            if (!isNonAssingmentOperator)
            {
                for (int a = 0; a < assignmentOperators.size(); a++)
                    if (line.COMPARE(i, assignmentOperators[a]->length(), *(assignmentOperators[a])) == 0)
                    {	    
                        if (assignmentOperators[a]->length() > 1)
                        {
                            outBuffer.append(assignmentOperators[a]->substr(1)); //<---
                            i += assignmentOperators[a]->length() - 1;
                        }
                        //cout<<"registering:" << assignmentOperators[a] <<'\n'; cout.flush();
                        if (!isInOperator && !isInTemplate)
                        {
                            registerInStatementIndent(line, i, spaceTabCount, isLineInStatement, false);
                            immediatelyPreviousAssignmentOp = assignmentOperators[a];
			    isInStatement = true;
                        }
                        break;  
                    }
            }
             
            if (isInOperator)
                isInOperator = false;
             
             
             
             
            //if ( parenDepth > 0 || !(isLegalNameChar(ch) || ch == ':'))
            //    isInStatement = true;
             
        }
         
        //cout << "end-tabCount=" << tabCount << "\n"; cout.flush();

        // handle special cases of unindentation:

        /*
         * if '{' doesn't follow an immediately previous '{' in the headerStack
         * (but rather another header such as "for" or "if", then unindent it
         * by one indentation relative to its block.
         */
        if (outBuffer.length()>0 && outBuffer[0]=='{'
                && !(headerStack->size()>1 && (*headerStack)[headerStack->size()-2] == &OPEN_BRACKET)
                && shouldIndentBrackettedLine)
            tabCount--;
         
        else if (outBuffer.length()>0 && outBuffer[0]=='}' && shouldIndentBrackettedLine )
            tabCount--;
         
        if (tabCount < 0)
            tabCount = 0;
         
        // take care of extra bracket indentatation option...
        if (bracketIndent && outBuffer.length()>0 && shouldIndentBrackettedLine)
            if (outBuffer[0]=='{' || outBuffer[0]=='}')
                tabCount++;
         
        // finally, insert indentations into begining of line
        for (i=0; i<tabCount; i++)
            outBuffer.insert(0, indentString);
         
        while ((spaceTabCount--) > 0)
            outBuffer.insert(0, " ");
         
        if (!inStatementIndentStack->empty())
        {
            if (statementTabCount < 0)
                statementTabCount = tabCount;
        }
        else
            statementTabCount = -1;
         
        return outBuffer;
    }

    void ASBeautifier::registerInStatementIndent(string &line, int i, int spaceTabCount, bool isLineInStatement, bool updateParenStack)
    {
        int inStatementIndent;
        int remainingCharNum = line.length() - i;
        int nextNonWSChar = 1;

        /*
        while (nextNonWSChar < remainingCharNum
               && (line.charAt(i+nextNonWSChar) == ' ' ||
        	   line.charAt(i+nextNonWSChar) == '\t') )
            nextNonWSChar++;
        */
        nextNonWSChar = getNextProgramCharDistance(line, i);

        // if indent is around the last char in the line, indent instead 2 spaces from the previous indent
        if (nextNonWSChar == remainingCharNum)	
        {
            int previousIndent = spaceTabCount;
            if (!inStatementIndentStack->empty())
                previousIndent = inStatementIndentStack->back();
             
            inStatementIndentStack->push_back(2 /*indentLength*/ + previousIndent );
            if (updateParenStack)
                parenIndentStack->push_back( previousIndent );
            return;
        }
        	
        if (updateParenStack)
            parenIndentStack->push_back(i+spaceTabCount);

        inStatementIndent = i + nextNonWSChar + spaceTabCount;

        if (i + nextNonWSChar > maxInStatementIndent)
            inStatementIndent =  indentLength*2 + spaceTabCount;
        	
        if (!inStatementIndentStack->empty() &&
                inStatementIndent < inStatementIndentStack->back())
            inStatementIndent = inStatementIndentStack->back();
        	
        //else if (!isLineInStatement && i + nextNonWSChar < 8)	
        //    inStatementIndent =  8 + spaceTabCount;

        inStatementIndentStack->push_back(inStatementIndent);
    }

    // get distance to the next non-white sspace, non-comment character in the line.
    // if no such character exists, return the length remaining to the end of the line.
    int ASBeautifier::getNextProgramCharDistance(string &line, int i)
    {
        int inStatementIndent;
        bool inComment = false;
        int remainingCharNum = line.length() - i;
        int charDistance = 1;
        int ch;

        for (charDistance = 1; charDistance < remainingCharNum; charDistance++)
        {
            ch = line[i + charDistance];
            if (inComment)
            {
                if (line.COMPARE(i + charDistance, 2, string(CLOSE_COMMENT)) == 0)
                {
                    charDistance++;
                    inComment = false;
                }
                continue;
            }
            else if (ch  == ' ' || ch == '\t')
                continue;
            else if (ch == '/')
            {
                if (line.COMPARE(i + charDistance, 2, string(OPEN_LINE_COMMENT)) == 0)
                    return remainingCharNum;
                else if (line.COMPARE(i + charDistance, 2, string(OPEN_COMMENT)) == 0)
                {
                    charDistance++;
                    inComment = true;
                }
            }
            else
                return charDistance;
        }

        return charDistance;
    }

    bool ASBeautifier::isLegalNameChar(char ch)
    {
        return ((ch>='a' && ch<='z') || (ch>='A' && ch<='Z') || (ch>='0' && ch<='9') ||
                ch=='.' || ch=='_' || (!isCStyle && ch=='$') || (isCStyle && ch=='~'));
    }

    int ASBeautifier::findLegalHeader(string &line, int i, vector<string*> &possibleHeaders)
    {
        int maxHeaders = possibleHeaders.size();
        int p;

        for (p=0; p < maxHeaders; p++)
            if (line.COMPARE(i, possibleHeaders[p]->length(), *(possibleHeaders[p])) == 0)
            {
                // first check that this is a header and not the begining of a longer word...
                int lineLength = line.length();
                int headerEnd = i + possibleHeaders[p]->length();
                char endCh = 0;
                	
                if ( headerEnd < lineLength )
                    endCh = line[headerEnd];
                if (headerEnd >= lineLength || !isLegalNameChar(endCh))
                    return p;
                else
                    return -1;
            }

        return -1;
    }


#ifdef USES_NAMESPACE
}
#endif
