/*
 * Copyright (c) 1998 Tal Davidson. All rights reserved.
 *
 * ASBeautifier.h 0.8.2 (part of astyle)
 * by Tal Davidson (davidsont@bigfoot.com)
 *
 * The astyle project and all files needed to compile it
 * are distributed under the "Artistic Licence" detailed below:
 *
 *
 *                           The ``Artistic License''
 *
 * Preamble
 * The intent of this document is to state the conditions under which a Package may
 * be copied, such that the Copyright Holder maintains some semblance of artistic
 * control over the development of the package, while giving the users of the
 * package the right to use and distribute the Package in a more-or-less customary
 * fashion, plus the right to make reasonable modifications.
 *
 * Definitions
 *     ``Package'' refers to the collection of files distributed by the Copyright
 *     Holder, and derivatives of that collection of files created through textual
 *     modification.
 *     ``Standard Version'' refers to such a Package if it has not been modified,
 *     or has been modified in accordance with the wishes of the Copyright Holder
 *     as specified below.
 *     ``Copyright Holder'' is whoever is named in the copyright or copyrights for
 *     the package.
 *     ``You'' is you, if you're thinking about copying or distributing this
 *     Package.
 *     ``Reasonable copying fee'' is whatever you can justify on the basis of media
 *     cost, duplication charges, time of people involved, and so on. (You will not
 *     be required to justify it to the Copyright Holder, but only to the computing
 *     community at large as a market that must bear the fee.)
 *     ``Freely Available'' means that no fee is charged for the item itself,
 *     though there may be fees involved in handling the item. It also means that
 *     recipients of the item may redistribute it under the same conditions they
 *     received it.
 *
 *     1. You may make and give away verbatim copies of the source form of the
 *        Standard Version of this Package without restriction, provided that you
 *        duplicate all of the original copyright notices and associated disclaimers.
 *     2. You may apply bug fixes, portability fixes and other modifications derived
 *        from the Public Domain or from the Copyright Holder. A Package modified in
 *        such a way shall still be considered the Standard Version.
 *     3. You may otherwise modify your copy of this Package in any way, provided that
 *        you insert a prominent notice in each changed file stating how and when you
 *        changed that file, and provided that you do at least ONE of the following:
 *         a. place your modifications in the Public Domain or otherwise make them
 *            Freely Available, such as by posting said modifications to Usenet or an
 *            equivalent medium, or placing the modifications on a major archive site
 *            such as uunet.uu.net, or by allowing the Copyright Holder to include
 *            your modifications in the Standard Version of the Package.
 *         b. use the modified Package only within your corporation or organization.
 *         c. rename any non-standard executables so the names do not conflict with
 *            standard executables, which must also be provided, and provide a
 *            separate manual page for each non-standard executable that clearly
 *            documents how it differs from the Standard Version.
 *         d. make other distribution arrangements with the Copyright Holder.
 *     4. You may distribute the programs of this Package in object code or executable
 *        form, provided that you do at least ONE of the following:
 *         a. distribute a Standard Version of the executables and library files,
 *            together with instructions (in the manual page or equivalent) on where
 *            to get the Standard Version.
 *         b. accompany the distribution with the machine-readable source of the
 *            Package with your modifications.
 *         c. give non-standard executables non-standard names, and clearly document
 *            the differences in manual pages (or equivalent), together with
 *            instructions on where to get the Standard Version.
 *         d. make other distribution arrangements with the Copyright Holder.
 *     5. You may charge a reasonable copying fee for any distribution of this
 *        Package. You may charge any fee you choose for support of this Package. You
 *        may not charge a fee for this Package itself. However, you may distribute
 *        this Package in aggregate with other (possibly commercial) programs as part
 *        of a larger (possibly commercial) software distribution provided that you do
 *        not advertise this Package as a product of your own. You may embed this
 *        Package's interpreter within an executable of yours (by linking); this shall
 *        be construed as a mere form of aggregation, provided that the complete
 *        Standard Version of the interpreter is so embedded.
 *     6. The scripts and library files supplied as input to or produced as output
 *        from the programs of this Package do not automatically fall under the
 *        copyright of this Package, but belong to whomever generated them, and may be
 *        sold commercially, and may be aggregated with this Package. If such scripts
 *        or library files are aggregated with this Package via the so-called "undump"
 *        or "unexec" methods of producing a binary executable image, then
 *        distribution of such an image shall neither be construed as a distribution
 *        of this Package nor shall it fall under the restrictions of Paragraphs 3 and
 *        4, provided that you do not represent such an executable image as a Standard
 *        Version of this Package.
 *     7. C subroutines (or comparably compiled subroutines in other languages)
 *        supplied by you and linked into this Package in order to emulate subroutines
 *        and variables of the language defined by this Package shall not be
 *        considered part of this Package, but are the equivalent of input as in
 *        Paragraph 6, provided these subroutines do not change the language in any
 *        way that would cause it to fail the regression tests for the language.
 *     8. Aggregation of this Package with a commercial distribution is always
 *        permitted provided that the use of this Package is embedded; that is, when
 *        no overt attempt is made to make this Package's interfaces visible to the
 *        end user of the commercial distribution. Such use shall not be construed as
 *        a distribution of this Package.
 *     9. The name of the Copyright Holder may not be used to endorse or promote
 *        products derived from this software without specific prior written
 *        permission.
 *        THIS PACKAGE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 *        WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 *        MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * The End
 */

#include <vector>
#include <string>

#include "compiler_defines.h"

#ifdef USES_NAMESPACE
using namespace std;

namespace astyle {
#endif

    class ASBeautifier
    {
        public:
            ASBeautifier();
            virtual ~ASBeautifier();
            virtual void init();
            void setTabIndentation();
            void setSpaceIndentation(int length=4);
            void setMaxInStatementIndentLength(int max);
            void setClassIndent(bool state);
            void setSwitchIndent(bool state);
            void setBracketIndent(bool state);
            void setCStyle();
            void setJavaStyle();
            virtual string beautify(string &line); 
             
        private:
            void initStatic();
            void registerInStatementIndent(string &line, int i, int spaceTabCount, bool isLineInStatement, bool updateParenStack);
            int getNextProgramCharDistance(string &line, int i);
            bool isLegalNameChar(char ch);
            int findLegalHeader(string &line, int i, vector<string*> &possibleHeaders);

            static vector<string*> headers;
            static vector<string*> nonParenHeaders;
            static vector<string*> preBlockStatements;
            static vector<string*> assignmentOperators;  
            static vector<string*> nonAssignmentOperators; 

            static string IF, ELSE;
            static string DO, WHILE;
            static string FOR;
            static string SWITCH, CASE, DEFAULT, OTHERWISE;
            static string TRY, CATCH, THROWS;
            static string PUBLIC, PROTECTED, PRIVATE;
            static string CLASS, STRUCT, UNION, INTERFACE, NAMESPACE;
            static string STATIC;
            static string SYNCHRONIZED;
            static string OPERATOR, TEMPLATE;
            static string OPEN_BRACKET, CLOSE_BRACKET;
            static string OPEN_LINE_COMMENT, OPEN_COMMENT, CLOSE_COMMENT;

            static string ASSIGN, PLUS_ASSIGN, MINUS_ASSIGN, MULT_ASSIGN;
            static string DIV_ASSIGN, MOD_ASSIGN, OR_ASSIGN, AND_ASSIGN;
            static string RETURN;
            static string EQUAL, PLUS_PLUS, MINUS_MINUS, NOT_EQUAL, GR_EQUAL, GR_GR;
            static string LS_EQUAL, LS_LS, ARROW, AND, OR;
             
            static bool calledInitStatic;

            // headerStack - a stack of the headers responsible for indentations of the current char
            vector<string*> *headerStack;
             
            // tempStacks - a stack of Stacks. Each inner stack holds the current header-list in a { } block.
            // The innermost { } block's stack sits at the top of the tempStacks.
            vector< vector<string*>* > *tempStacks;
             
            // blockParenDepthStack - stack of the number of parenthesies that are open when new NESTED BLOCKS are created. 
            vector<int> *blockParenDepthStack;
             
            // blockStatementStack - stack of the states of 'isInStatement' when new NESTED BLOCKS are created. 
            vector<bool>  *blockStatementStack;
             
            // parenStatementStack - stack of the states of 'isInStatement' when new NESTED PARENTHESIES are created.  
            vector<bool> *parenStatementStack;
             
            // inStatementIndentStack - stack of LOCATIONS of in-statement indents
            vector<int> *inStatementIndentStack;
             
            // inStatementIndentStackSizeStack - stack of SIZES of inStatementIndentStack stacks
            vector<int> *inStatementIndentStackSizeStack;
             
            // parenIndentStack - stack of LOCATIONS of '(' or '[' chars
            vector<int> *parenIndentStack;
             
            // bracketBlockStateStack - stack of types of nested '{' brackets.
            // Each element of the stack is either True (=the beginner of a block), or False (=the beginner of a
            // static array).  
            vector<bool> *bracketBlockStateStack;
             
            // isSpecialChar - true if a there exists a '\' preceding the current chararacter.
            //   i.e. \n, \t, \\, ...
            //bool isSpecialChar;
             
            // isInQuote - true when the current character is part of a quote (i.e. 'g' or "ffff")
            bool isInQuote;
             
            // isInComment - true when current character is part of a /* */ comment
            bool isInComment;
             
            // isInCase - true if in middle of a case statement (inside a switch);
            bool isInCase;
             
            // isInQuestion - true if in the middle of a '? :' statement
            bool isInQuestion;
             
            // isInStatement - true when current character is a part of an ongoing statement
            bool isInStatement;
             
            // isInClassHeader - true if inside a 'class' statement
            bool isInClassHeader;
             
            // isInClassHeaderTab - true if a special tab has been activated for the 'class statement'
            bool isInClassHeaderTab;  
             
            // switchIndent - true if switch blocks should have an additional internal indent.
            bool switchIndent;
             
            // bracketIndent - true if brackets should have an added indent.
            bool bracketIndent;
             
            // quoteChar - the quote delimeter of a quote (' or ")
            char quoteChar;
             
            // parenDepth - the depth of parenthesies around the current character
            int parenDepth;
             
            // indentString - the String to be used for every indentation
            // - either a "\t" or a String of n spaces.
            string indentString;
             
            // indentLength - the length of one indent unit.
            int indentLength;
             
            // blockTabCount - stores number of tabs to add to begining of line
            // due to statements with INNER blocks inside open parenthesies.
            int blockTabCount;  
             
            int statementTabCount;
            int leadingWhiteSpaces;
            int maxInStatementIndent;
            char prevNonSpaceCh;
            char currentNonSpaceCh;
            string *currentHeader;
            bool isInHeader;
            string *immediatelyPreviousAssignmentOp;
            bool isCStyle;
            bool isInOperator;
            bool isInTemplate;
            bool classIndent;
             
            char currentNonLegalCh;
            char prevNonLegalCh;
            int templateDepth;  
    };


#ifdef USES_NAMESPACE
}
#endif
