/*
 * Copyright (c) 1998 Tal Davidson. All rights reserved.
 *
 * ASBeautifier_main.cpp 0.8.2 (part of astyle)
 * by Tal Davidson (davidsont@bigfoot.com)
 *
 * The astyle project and all files needed to compile it
 * are distributed under the "Artistic Licence" detailed below:
 *
 *
 *                           The ``Artistic License''
 *
 * Preamble
 * The intent of this document is to state the conditions under which a Package may
 * be copied, such that the Copyright Holder maintains some semblance of artistic
 * control over the development of the package, while giving the users of the
 * package the right to use and distribute the Package in a more-or-less customary
 * fashion, plus the right to make reasonable modifications.
 *
 * Definitions
 *     ``Package'' refers to the collection of files distributed by the Copyright
 *     Holder, and derivatives of that collection of files created through textual
 *     modification.
 *     ``Standard Version'' refers to such a Package if it has not been modified,
 *     or has been modified in accordance with the wishes of the Copyright Holder
 *     as specified below.
 *     ``Copyright Holder'' is whoever is named in the copyright or copyrights for
 *     the package.
 *     ``You'' is you, if you're thinking about copying or distributing this
 *     Package.
 *     ``Reasonable copying fee'' is whatever you can justify on the basis of media
 *     cost, duplication charges, time of people involved, and so on. (You will not
 *     be required to justify it to the Copyright Holder, but only to the computing
 *     community at large as a market that must bear the fee.)
 *     ``Freely Available'' means that no fee is charged for the item itself,
 *     though there may be fees involved in handling the item. It also means that
 *     recipients of the item may redistribute it under the same conditions they
 *     received it.
 *
 *     1. You may make and give away verbatim copies of the source form of the
 *        Standard Version of this Package without restriction, provided that you
 *        duplicate all of the original copyright notices and associated disclaimers.
 *     2. You may apply bug fixes, portability fixes and other modifications derived
 *        from the Public Domain or from the Copyright Holder. A Package modified in
 *        such a way shall still be considered the Standard Version.
 *     3. You may otherwise modify your copy of this Package in any way, provided that
 *        you insert a prominent notice in each changed file stating how and when you
 *        changed that file, and provided that you do at least ONE of the following:
 *         a. place your modifications in the Public Domain or otherwise make them
 *            Freely Available, such as by posting said modifications to Usenet or an
 *            equivalent medium, or placing the modifications on a major archive site
 *            such as uunet.uu.net, or by allowing the Copyright Holder to include
 *            your modifications in the Standard Version of the Package.
 *         b. use the modified Package only within your corporation or organization.
 *         c. rename any non-standard executables so the names do not conflict with
 *            standard executables, which must also be provided, and provide a
 *            separate manual page for each non-standard executable that clearly
 *            documents how it differs from the Standard Version.
 *         d. make other distribution arrangements with the Copyright Holder.
 *     4. You may distribute the programs of this Package in object code or executable
 *        form, provided that you do at least ONE of the following:
 *         a. distribute a Standard Version of the executables and library files,
 *            together with instructions (in the manual page or equivalent) on where
 *            to get the Standard Version.
 *         b. accompany the distribution with the machine-readable source of the
 *            Package with your modifications.
 *         c. give non-standard executables non-standard names, and clearly document
 *            the differences in manual pages (or equivalent), together with
 *            instructions on where to get the Standard Version.
 *         d. make other distribution arrangements with the Copyright Holder.
 *     5. You may charge a reasonable copying fee for any distribution of this
 *        Package. You may charge any fee you choose for support of this Package. You
 *        may not charge a fee for this Package itself. However, you may distribute
 *        this Package in aggregate with other (possibly commercial) programs as part
 *        of a larger (possibly commercial) software distribution provided that you do
 *        not advertise this Package as a product of your own. You may embed this
 *        Package's interpreter within an executable of yours (by linking); this shall
 *        be construed as a mere form of aggregation, provided that the complete
 *        Standard Version of the interpreter is so embedded.
 *     6. The scripts and library files supplied as input to or produced as output
 *        from the programs of this Package do not automatically fall under the
 *        copyright of this Package, but belong to whomever generated them, and may be
 *        sold commercially, and may be aggregated with this Package. If such scripts
 *        or library files are aggregated with this Package via the so-called "undump"
 *        or "unexec" methods of producing a binary executable image, then
 *        distribution of such an image shall neither be construed as a distribution
 *        of this Package nor shall it fall under the restrictions of Paragraphs 3 and
 *        4, provided that you do not represent such an executable image as a Standard
 *        Version of this Package.
 *     7. C subroutines (or comparably compiled subroutines in other languages)
 *        supplied by you and linked into this Package in order to emulate subroutines
 *        and variables of the language defined by this Package shall not be
 *        considered part of this Package, but are the equivalent of input as in
 *        Paragraph 6, provided these subroutines do not change the language in any
 *        way that would cause it to fail the regression tests for the language.
 *     8. Aggregation of this Package with a commercial distribution is always
 *        permitted provided that the use of this Package is embedded; that is, when
 *        no overt attempt is made to make this Package's interfaces visible to the
 *        end user of the commercial distribution. Such use shall not be construed as
 *        a distribution of this Package.
 *     9. The name of the Copyright Holder may not be used to endorse or promote
 *        products derived from this software without specific prior written
 *        permission.
 *        THIS PACKAGE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 *        WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 *        MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * The End
 */


#include <iostream>
#include <fstream>
#include <string>
#include "compiler_defines.h"
#include "ASBeautifier.h"


#ifdef USES_NAMESPACE
using namespace std;
using namespace astyle;
#endif


void beautifyStream(ASBeautifier beautifier, istream &in, ostream &out)
{
  char buf[400];
  while (in)
    {
      in.getline(buf, 399);
      out << beautifier.beautify(string(buf));
      out << '\n';
    }
}


main(int argc, char *argv[])
{
  ASBeautifier beautifier;;
  vector<string> fileNameVector;
  bool isHelpShown = false;
  
  // manage flags
  for (int i=1; i<argc; i++)
    {
      string arg = string(argv[i]);
      if (arg.compare("-j") == 0)
	beautifier.setJavaStyle();
      else if (arg.compare("-t") == 0)
	beautifier.setTabIndentation();
      else if (arg.COMPARE(0, 2, string("-s")) == 0)
	{
	  int spaceNum = 4;
	  if (arg.length() > 2)
	    spaceNum = atoi(arg.substr(2).data());	  
	    
	  beautifier.setSpaceIndentation(spaceNum);
	}
      else if (arg.COMPARE(0, 2, string("-m")) == 0)
	{
	  int maxIndent = 40;
	  if (arg.length() > 2)
	    maxIndent = atoi(arg.substr(2).data());      
	  beautifier.setMaxInStatementIndentLength(maxIndent);
	}
      else if (arg.compare("-ib") == 0)
	beautifier.setBracketIndent(true);
      else if (arg.compare("-fs") == 0)
	beautifier.setSwitchIndent(false);            
      else if (arg.COMPARE(0, 1, string("-")) == 0 && !isHelpShown)
	{
	  isHelpShown = true;
	  
	  cerr << "\n";
	  cout << "ASBeautifier 0.8.2   (created by Tal Davidson, davidsont@bigfoot.com)\n";
	  cerr << "\n";
	  cerr << "Usage  :  astyle [options] < Original > Beautified\n";
	  cerr << "          astyle [options] Foo.cpp Bar.cpp  [...]\n";
	  cerr << "\n";
	  cerr << "When given a specific file, astyle will create an output file with a\n";                
	  cerr << "suffix of \".as\" added to the original filename, i.e: Foo.java --> Foo.java.as\n";
	  cerr << "\n";
	  cerr << "By default, ASBeautifier is set up to indent C/C++  files, with 4 spaces per\n" ;
	  cerr << "indent, and a maximal indentation of 40 spaces inside continuous statements.\n";
	  cerr << "\n";
	  cerr << "Options:  -j   Indent a Java(TM) source file\n";
	  cerr << "          -t   Indent using tab characters\n";
	  cerr << "          -s#  Indent using # spaces per indent (i.e. -s4)\n" ;
	  cerr << "          -m#  Indent a maximal # spaces in a continuous statement,\n";
	  cerr << "               relatively to the previous line(i.e. -m40)\n";
	  cerr << "          -ib  add extra indentation to brackets\n";
	  cerr << "          -fs  flush (i.e. don't indent) 'switch' blocks\n";
	  cerr << "          -h   Print this help message\n";
	  cerr << "\n";
	  cerr << "HomePage: http://www.bigfoot.com/~davidsont/astyle/ASBeautifier.html\n";
	  cerr << "\n";

	  return 0;                
	}
      else // file-name
	fileNameVector.push_back(arg);
      
    }
  
  if (fileNameVector.empty())
    {
      beautifyStream(beautifier, cin, cout);
      cout.flush();      
    }
  else
    {
      for (int i=0; i<fileNameVector.size(); i++)
	{
	  beautifier.init();
	  //try {
	  string &fileName = fileNameVector[i];
	  ifstream in(fileName.data());
	  ofstream out(string(fileName).append(".as").data());
	  
	  beautifyStream(beautifier, in, out);
	  //}
	  //catch (IOException e) {
	  //    System.err.println("Error: " + e);
	  //  }
	  
	  out.flush();
	  out.close();
	  
	  in.close();
	}
    }
	return 1;
}




